"""Control Linear Garage Doors."""

from __future__ import annotations

import asyncio
import logging
import uuid
from typing import Any, Awaitable, Callable

import aiohttp

from ._util import create_request
from .const import MessageTypes
from .errors import (
    InvalidDeviceIDError,
    InvalidLoginError,
    NotOpenError,
    UnexpectedError,
)
from .ws import WebSocketMonitor


class Linear:
    """A Linear account.

    Instantiate this class then run [`login()`][linear_garage_door.Linear.login] with your credentials to connect.

    Example:
        ```python
        from linear_garage_door import Linear

        async def on_device_state_event(data: dict[str, dict[str, str] | str]):
            print(data)

        linear = Linear(on_device_state_event=on_device_state_event)

        # Log in to the account
        await linear.login("email@email.com", "password", "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx")

        # Get the sites on the account
        sites = await linear.get_sites()

        # Loop through each site and get the devices
        for site in sites:
            devices = await linear.get_devices(site["id"])
            for device in devices:
                print("")
                print(device["name"], "----")
                device_state = await linear.get_device_state(device["id"])
                for key in device_state.keys():
                    print(key, "--")
                    print("")
                    for attribute in device_state[key].keys():
                        print(attribute, "=", device_state[key][attribute])
                    print("")

        await linear.close()
        ```
    """

    _message_id = -1
    _ws: aiohttp.ClientWebSocketResponse
    _device_id: str
    _logger: logging.Logger
    _debug: bool = False
    _user_id: str
    _user_email: str
    _internal_callback: Callable[[dict[str, Any]], Awaitable[None]] | None = None
    _monitor: WebSocketMonitor

    def __init__(
        self: Linear,
        on_device_state_event: Callable[[dict[str, Any]], Awaitable[None]]
        | None = None,
    ):
        """Initialize the Linear account.

        Args:
            on_device_state_event (function | None, optional): Pass a function to subscribe to device state events. The function provided will be called with the data similar to the data returned by [`login()`][linear_garage_door.Linear.login]. Defaults to None.
        """
        self._logger = logging.getLogger("linear_garage_door")
        self._debug = self._logger.isEnabledFor(logging.DEBUG)
        self.on_device_state_event = on_device_state_event

    @property
    def open(self: Linear) -> bool:
        return not self._ws.closed

    def _get_message_id(self: Linear) -> str:
        self._message_id += 1
        message_id = hex(self._message_id)[2:].zfill(8).upper()

        return message_id

    def _verify_device_id(self: Linear, device_id: str | None) -> str:
        if device_id is None:
            device_id = str(uuid.uuid4()).upper()
            self._logger.warning(
                "No device_id provided. Proceeding with random UUID. Please use this UUID in future calls to this function: %s",
                device_id,
            )
        else:
            valid = True
            try:
                uuid.UUID(str(device_id))
            except ValueError:
                valid = False

            if not valid:
                raise InvalidDeviceIDError("device_id must be a UUID.")
            else:
                device_id = device_id.upper()

        return device_id

    async def _await_for_any_msg(self: Linear) -> dict[str, Any]:
        future: asyncio.Future[dict[str, Any]] = asyncio.Future()

        async def _got_resp(data: dict[str, Any]) -> None:
            future.set_result(data)
            self._internal_callback = None

        self._internal_callback = _got_resp

        return await future

    async def _await_for_msg_type(self: Linear, msg_type: str) -> dict[str, Any]:
        future: asyncio.Future[dict[str, Any]] = asyncio.Future()

        async def _got_resp(data: dict[str, Any]) -> None:
            if data["Type"] == msg_type:
                future.set_result(data)
                self._internal_callback = None

        self._internal_callback = _got_resp

        return await future

    async def _await_for_dev_state_report(self: Linear) -> dict[str, Any]:
        future: asyncio.Future[dict[str, Any]] = asyncio.Future()

        async def _got_resp(response: dict[str, Any]) -> None:
            if (
                response["Type"] == MessageTypes.DEVICE_STATE.value
                and response["Headers"]["Type"] == "Report"
            ):
                future.set_result(response)
                self._internal_callback = None

        self._internal_callback = _got_resp

        return await future

    async def _message_callback(self: Linear, data: dict[str, Any]) -> None:
        if self._internal_callback:
            # Override user-provided callback
            await self._internal_callback(data)
        elif self.on_device_state_event:
            await self.on_device_state_event(data)

    def _check_if_open(self: Linear) -> None:
        if not self.open:
            raise NotOpenError("The WebSocket has not been opened. Call login() first.")

    async def login(
        self: Linear,
        email: str,
        password: str,
        device_id: str | None = None,
        client_session: aiohttp.ClientSession | None = None,
    ) -> dict[str, Any] | None:
        """Logs in to a Linear account.

        Tip:
            Need a device ID? Grab one [here](https://www.uuidgenerator.net/version4).

        Example:
            ```python
            from linear_garage_door import Linear

            linear = Linear()
            await linear.login("email@email.com", "password", "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx")

            await linear.close()
            ```

        Args:
            email (str): The email of the account to log in to.
            password (str): The password of the account to log in to.
            device_id (str | None, optional): The device ID of your choice. Must be a UUID. Defaults to a random UUID.

        Raises:
            InvalidLogin: The login provided is invalid.
            InvalidDeviceID: The device ID provided is invalid.

        Returns:
            A dict with the type of response, most likely "WELCOME", the headers of the response, and the body.
        """

        try:
            if self.open:
                return None
        except AttributeError:
            pass

        self._device_id = "M/" + self._verify_device_id(device_id)
        request = create_request(
            MessageTypes.HELLO,
            {
                "MessageID": self._get_message_id(),
                "UserEmail": email,
                "Password": password,
                "TargetsHere": self._device_id,
                "Targets": "B/*",
                "SendingDeviceID": self._device_id,
            },
        )

        ws_monitor = WebSocketMonitor()
        if ws_monitor.websocket is None or ws_monitor.websocket.closed:
            if client_session is None:
                await ws_monitor.new_connection(
                    aiohttp.ClientSession(), self._message_callback, True, False
                )
            else:
                await ws_monitor.new_connection(
                    client_session, self._message_callback, True, True
                )
        if ws_monitor.monitor is None or ws_monitor.monitor.done():
            await ws_monitor.start_monitor()
        ws = ws_monitor.websocket

        if ws is None:
            raise UnexpectedError("ws is None when it should be.")

        self._monitor = ws_monitor
        self._ws = ws

        await ws.send_str(request)

        response = await self._await_for_any_msg()

        if (
            response["Type"] == MessageTypes.GOODBYE.value
            and response["Headers"]["UserID"] == ""
        ):
            if response["Headers"]["Reason"] == "InvalidLogin":
                raise InvalidLoginError(
                    "Login provided is invalid, please check the email and password"
                )

            raise InvalidLoginError("Login error: ", response["Headers"]["Reason"])

        self._user_id = response["Headers"]["UserID"]
        self._user_email = response["Headers"]["UserEmail"]

        return response

    async def get_sites(self: Linear) -> list[dict[str, str]]:
        """Get sites available under this account.

        Example:
            ```python
            from linear_garage_door import Linear

            linear = Linear()
            await linear.login("email@email.com", "password", "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx")

            sites = await linear.get_sites()
            for site in sites:
                print(site["name"], "has id", site["id"])

            await linear.close()
            ```

        Raises:
            NotOpen: The WebSocket has not been opened.

        Returns:
            A list of dicts that contain the device's ID and name.
        """

        self._check_if_open()

        request = create_request(
            MessageTypes.REQUEST_SITE_LIST,
            {
                "MessageID": self._get_message_id(),
                "SendingUserID": self._user_id,
                "Targets": "B/*",
                "SendingDeviceID": self._device_id,
            },
        )

        await self._ws.send_str(request)
        response = await self._await_for_msg_type(MessageTypes.SITE_LIST.value)

        sites = []

        for key in response["Headers"].keys():
            if key.startswith("Site-"):
                site = response["Headers"][key].split(",")
                sites.append({"id": site[0], "name": site[1]})

        return sites

    async def get_devices(self: Linear, site: str) -> list[dict[str, list[str] | str]]:
        """Get devices available under a specific site.

        Example:
            ```python
            from linear_garage_door import Linear

            linear = Linear()
            await linear.login("email@email.com", "password", "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx")

            sites = await linear.get_sites()
            for site in sites:
                devices = await linear.get_devices(site["id"])
                for device in devices:
                    print(device["name"], "(", device["id"], ") has the following subdevices:")
                    for subdevice in device["subdevices"]:
                        print(subdevice)

            await linear.close()
            ```

        Args:
            site (str): The site ID to fetch the devices from. Get site IDs by calling [`get_sites()`][linear_garage_door.Linear.get_sites].

        Raises:
            NotOpen: The WebSocket has not been opened.

        Returns:
            A list of dictionaries that contains the ID, name, and subdevices of that device. (Subdevices usually are GDO, Light, etc...)
        """

        self._check_if_open()

        request = create_request(
            MessageTypes.REQUEST_SITE_CONFIG,
            {
                "MessageID": self._get_message_id(),
                "SendingUserID": self._user_id,
                "Targets": "B/*",
                "SiteID": site,
                "SendingDeviceID": self._device_id,
            },
        )

        await self._ws.send_str(request)
        response = await self._await_for_msg_type(MessageTypes.SITE_CONFIG.value)

        devices = []

        for key in response["Headers"].keys():
            if key.startswith("HDev-"):
                device = response["Headers"][key].split(",")
                devices.append(
                    {
                        "id": device[0],
                        "name": device[1],
                        "subdevices": [i.replace("\\", "") for i in device[5:]],
                    }
                )

        return devices

    async def operate_device(
        self: Linear, device_id: str, subdevice: str, subdevice_state: str
    ) -> None:
        """Operate a device.

        Here are the known states for each subdevice:

        - `GDO`: `Open` or `Close` (Not `Closed`!)
        - `Light`: `On` or `Off`

        More reverse engineering will be required in the future for setting the Light brightness and getting the progress of door opening.

        This does not return anything. Events will be fired from the server and can give you a state.

        Example:
            ```python
            from linear_garage_door import Linear

            linear = Linear()

            await linear.login("email@email.com", "password", "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx")

            await linear.operate_device(device["id"], "GDO", "Open")
            ```

        Args:
            device_id (str): The device ID you want to operate. You can get this from [`get_devices()`][linear_garage_door.Linear.get_devices]
            subdevice (str): The subdevice you want to operate. The list of subdevices is also available from [`get_devices()`][linear_garage_door.Linear.get_devices]
            subdevice_state (str): The state to set the subdevice to. I don't know where the full list of states are, but you can get a good guess from [`get_device_state()`][linear_garage_door.Linear.get_device_state]

        Raises:
            NotOpen: The WebSocket has not been opened.
        """

        self._check_if_open()

        # Send HELLO request to device

        hello_request = create_request(
            MessageTypes.HELLO,
            {
                "MessageID": self._get_message_id(),
                "SendingUserID": self._user_id.upper(),
                "UserEmail": self._user_email,
                "TargetsHere": self._device_id,
                "Targets": device_id,
                "SendingDeviceID": self._device_id,
            },
        )

        await self._ws.send_str(hello_request)

        op_request = create_request(
            MessageTypes.OPERATE_DEVICE,
            {
                "MessageID": self._get_message_id(),
                "SendingUserID": self._user_id.upper(),
                "SubDev-" + subdevice: subdevice_state,
                "Targets": "H/" + device_id,
                "SendingDeviceID": self._device_id,
            },
        )

        await self._ws.send_str(op_request)

        await self._await_for_msg_type(MessageTypes.DEVICE_STATE.value)

        return

    async def get_device_state(
        self: Linear, device_id: str
    ) -> dict[str, dict[str, str]]:
        """Get state of device.

        Example:
            ```python
            from linear_garage_door import Linear

            async def on_device_state_event(data: dict[str, dict[str, str] | str]):
                print(data)

            linear = Linear(on_device_state_event=on_device_state_event)

            # Log in to the account
            await linear.login("email@email.com", "password", "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx")

            # Get the sites on the account
            sites = await linear.get_sites()

            # Loop through each site and get the devices
            for site in sites:
                devices = await linear.get_devices(site["id"])
                for device in devices:
                    print("")
                    print(device["name"], "----")
                    device_state = await linear.get_device_state(device["id"])
                    for key in device_state.keys():
                        print(key, "--")
                        print("")
                        for attribute in device_state[key].keys():
                            print(attribute, "=", device_state[key][attribute])
                        print("")

            await linear.close()
            ```

        Args:
            device_id (str): The device ID to get the state of.

        Raises:
            NotOpen: The WebSocket has not been opened.

        Returns:
            A dictionary that shows each subdevice and their respective states.
        """

        self._check_if_open()

        # Send HELLO request to device

        hello_request = create_request(
            MessageTypes.HELLO,
            {
                "MessageID": self._get_message_id(),
                "SendingUserID": self._user_id.upper(),
                "UserEmail": self._user_email,
                "TargetsHere": self._device_id,
                "Targets": device_id,
                "SendingDeviceID": self._device_id,
            },
        )

        await self._ws.send_str(hello_request)

        state_request = create_request(
            MessageTypes.REQUEST_DEVICE_STATE,
            {
                "MessageID": self._get_message_id(),
                "SendingUserID": self._user_id.upper(),
                "Targets": "H/" + device_id,
                "SendingDeviceID": self._device_id,
            },
        )

        await self._ws.send_str(state_request)

        response = await self._await_for_dev_state_report()

        subdev_props: dict[str, dict[str, str]] = {}

        for key in response["Headers"].keys():
            if key.startswith("SubDev-"):
                subdev_name = key.split("SubDev-")[1]
                subdev_props[subdev_name] = {}

                subdev = response["Headers"][key].split(",")

                for prop in subdev:
                    subdev_props[subdev_name][prop.split(":")[0]] = prop.split(":")[1]

        return subdev_props

    async def close(self: Linear) -> None:
        """Close the WebSocket."""
        await self._monitor.close()



from __future__ import annotations

from typing import Any

from .const import TYPE_TO_PROTOCOL, MessageTypes


def parse_response(resp: str) -> dict[str, Any]:
    response_dict: dict[str, Any] = {
        "Type": "",
        "Headers": {},
        "Body": "",
    }
    response = resp.split("\r\n")

    for i in range(1, len(response)):
        line = response[i]
        if line == "":
            break
        response_dict["Headers"][line.split(":")[0]] = ":".join(line.split(":")[1:])

    response_dict["Type"] = response[0].split(" ")[1]

    return response_dict


def create_request(msg_type: MessageTypes, data: dict[str, str]) -> str:
    request_string = f"{TYPE_TO_PROTOCOL[msg_type]} {msg_type.value}\r\n"

    for key in data.keys():
        request_string += f"{key}:{data[key]}\r\n"

    request_string += "\r\n"

    return request_string



class InvalidLoginError(Exception):
    """The login provided is invalid."""


class InvalidDeviceIDError(Exception):
    """The device ID provided is invalid."""


class NotOpenError(Exception):
    """The WebSocket has not been opened."""


class UnexpectedError(Exception):
    """This error should never be raised."""



from __future__ import annotations

import asyncio
from random import uniform
from typing import Any, Awaitable, Callable

from aiohttp import ClientSession, ClientWebSocketResponse, WSMsgType

from ._util import parse_response
from .const import SERVICE_URL


async def cancel_task(*tasks: asyncio.Task[Any] | None) -> None:
    for task in tasks:
        if task is not None and not task.done():
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass


class WebSocketMonitor:
    def __init__(self) -> None:
        self._disconnect = False
        self._ws: ClientWebSocketResponse | None = None
        self._monitor_task: asyncio.Task[Any] | None = None
        self._receiver_task: asyncio.Task[Any] | None = None
        self._callback: Callable[[dict[str, Any]], Awaitable[None]] | None = None
        self._session: ClientSession

    @property
    def connected(self) -> bool:
        if self._disconnect:
            return False
        return False if self._ws is None else not self._ws.closed

    @property
    def websocket(self) -> ClientWebSocketResponse | None:
        return self._ws

    @property
    def monitor(self) -> asyncio.Task[Any] | None:
        return self._monitor_task

    async def new_connection(
        self,
        session: ClientSession,
        callback: Callable[[dict[str, Any]], Awaitable[None]] | None = None,
        start_monitor: bool = False,
        custom_session: bool = False,
    ) -> None:
        await cancel_task(self._receiver_task)
        self._disconnect = False
        self._ws = await session.ws_connect(SERVICE_URL, ssl=False)
        self._receiver_task = asyncio.ensure_future(self._receiver())
        self._callback = callback
        self._session = session
        self._custom_session = custom_session
        if start_monitor:
            await self.start_monitor()

    async def _receiver(self) -> None:
        if not (websocket := self._ws):
            return
        while not websocket.closed:
            msg = await websocket.receive()
            if msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSING, WSMsgType.CLOSED):
                break

            if msg.type == WSMsgType.TEXT:
                response = parse_response(msg.data)
                if self._callback is not None:
                    await self._callback(response)

    async def _monitor(self) -> None:
        attempt = 0
        while not self._disconnect:
            while self.connected:
                await asyncio.sleep(1)
            if not self._disconnect:
                await self.new_connection(self._session, self._callback)
            if not self._ws or self._ws.closed:
                await asyncio.sleep(min(1 * 2**attempt + uniform(0, 1), 300))
                attempt += 1
                continue
            attempt = 0

    async def start_monitor(self) -> None:
        if self._monitor_task is None or self._monitor_task.done():
            self._monitor_task = asyncio.ensure_future(self._monitor())

    async def stop_monitor(self) -> None:
        await cancel_task(self._monitor_task)

    async def close(self) -> None:
        self._disconnect = True
        if self._ws:
            await self._ws.close()
            if not self._custom_session:
                await self._session.close()
        await cancel_task(self._monitor_task, self._receiver_task)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['linear_garage_door']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'certifi>=2022.12.7,<2023.0.0',
 'dnspython>=2.3.0,<3.0.0',
 'websocket-client>=1.5.0,<2.0.0']

setup_kwargs = {
    'name': 'linear-garage-door',
    'version': '0.2.2',
    'description': 'Control Linear Garage Doors with Python',
    'long_description': "# Linear Garage Door\n\n[![PyPI](https://img.shields.io/pypi/v/linear-garage-door?style=flat-square)](https://pypi.python.org/pypi/linear-garage-door/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/linear-garage-door?style=flat-square)](https://pypi.python.org/pypi/linear-garage-door/)\n[![PyPI - License](https://img.shields.io/pypi/l/linear-garage-door?style=flat-square)](https://pypi.python.org/pypi/linear-garage-door/)\n[![Coookiecutter - Wolt](https://img.shields.io/badge/cookiecutter-Wolt-00c2e8?style=flat-square&logo=cookiecutter&logoColor=D4AA00&link=https://github.com/woltapp/wolt-python-package-cookiecutter)](https://github.com/woltapp/wolt-python-package-cookiecutter)\n\n\n---\n\n**Documentation**: [https://IceBotYT.github.io/linear-garage-door](https://IceBotYT.github.io/linear-garage-door)\n\n**Source Code**: [https://github.com/IceBotYT/linear-garage-door](https://github.com/IceBotYT/linear-garage-door)\n\n**PyPI**: [https://pypi.org/project/linear-garage-door/](https://pypi.org/project/linear-garage-door/)\n\n---\n\nControl Linear Garage Doors with Python\n\n## Installation\n\n```sh\npip install linear-garage-door\n```\n\n## Development\n\n* Clone this repository\n* Requirements:\n  * [Poetry](https://python-poetry.org/)\n  * Python 3.7+\n* Create a virtual environment and install the dependencies\n\n```sh\npoetry install\n```\n\n* Activate the virtual environment\n\n```sh\npoetry shell\n```\n\n### Testing\n\n```sh\npytest\n```\n\n### Documentation\n\nThe documentation is automatically generated from the content of the [docs directory](./docs) and from the docstrings\n of the public signatures of the source code. The documentation is updated and published as a [Github project page\n ](https://pages.github.com/) automatically as part each release.\n\n### Releasing\n\nTrigger the [Draft release workflow](https://github.com/IceBotYT/linear-garage-door/actions/workflows/draft_release.yml)\n(press _Run workflow_). This will update the changelog & version and create a GitHub release which is in _Draft_ state.\n\nFind the draft release from the\n[GitHub releases](https://github.com/IceBotYT/linear-garage-door/releases) and publish it. When\n a release is published, it'll trigger [release](https://github.com/IceBotYT/linear-garage-door/blob/master/.github/workflows/release.yml) workflow which creates PyPI\n release and deploys updated documentation.\n\n### Pre-commit\n\nPre-commit hooks run all the auto-formatters (e.g. `black`, `isort`), linters (e.g. `mypy`, `flake8`), and other quality\n checks to make sure the changeset is in good shape before a commit/push happens.\n\nYou can install the hooks with (runs for each commit):\n\n```sh\npre-commit install\n```\n\nOr if you want them to run only for each push:\n\n```sh\npre-commit install -t pre-push\n```\n\nOr if you want e.g. want to run all checks manually for all files:\n\n```sh\npre-commit run --all-files\n```\n\n---\n\nThis project was generated using the [wolt-python-package-cookiecutter](https://github.com/woltapp/wolt-python-package-cookiecutter) template.\n",
    'author': 'IceBotYT',
    'author_email': 'icebotyt@outlook.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://IceBotYT.github.io/linear-garage-door',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from enum import Enum

SERVICE_URL = "wss://linr-cs-tm-prod.trafficmanager.net:8080/Message.svc"  # One URL for everything


class MessageTypes(Enum):
    HELLO = "HELLO"
    KEEPALIVE = "keepalive"
    REQUEST_SITE_LIST = "REQUEST_SITE_LIST"
    REQUEST_DEVICE_STATE = "REQUEST_DEVICE_STATE"
    REQUEST_HISTORY = "REQUEST_HISTORY"
    SITE_LIST = "SITE_LIST"
    GOODBYE = "GOODBYE"
    REQUEST_SITE_CONFIG = "REQUEST_SITE_CONFIG"
    SITE_CONFIG = "SITE_CONFIG"
    OPERATE_DEVICE = "OPERATE_DEVICE"
    DEVICE_STATE = "DEVICE_STATE"
    WELCOME = "WELCOME"


TYPE_TO_PROTOCOL = {
    MessageTypes.HELLO: "CMP/1.0",
    MessageTypes.KEEPALIVE: "ECHO/1.0",
    MessageTypes.REQUEST_SITE_LIST: "SMP/1.0",
    MessageTypes.REQUEST_DEVICE_STATE: "ACP/1.0",
    MessageTypes.REQUEST_HISTORY: "ERP/1.0",
    MessageTypes.REQUEST_SITE_CONFIG: "SMP/1.0",
    MessageTypes.OPERATE_DEVICE: "ACP/1.0",
    MessageTypes.DEVICE_STATE: "ACP/1.0",
}
